

import org.apache.cordova.CordovaPlugin;
import org.apache.cordova.CallbackContext;
import org.apache.cordova.PluginResult;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import android.content.Intent;
import android.net.Uri;
import android.os.Bundle;
import android.util.Log;
import java.util.ArrayList;

import com.spotify.android.appremote.api.ConnectionParams;
import com.spotify.android.appremote.api.Connector;
import com.spotify.android.appremote.api.SpotifyAppRemote;

import com.spotify.protocol.types.PlayerState;
import com.spotify.protocol.types.Track;
import com.spotify.sdk.android.auth.AuthorizationClient;
import com.spotify.sdk.android.auth.AuthorizationRequest;
import com.spotify.sdk.android.auth.AuthorizationResponse;


public class CordovaSpotifyRemote extends CordovaPlugin {

    private static final int REQUEST_CODE = 1337;
    private static final String REDIRECT_URI = "latitunes://callback";

    private static final int PLAY = 3;
    private static final int PAUSE = 4;
    private static final int RESUME = 5;
    private static final int SKIPNEXT = 6;
    private static final int SKIPBACK = 7;
    private static final int PLAYQUEUE = 8;
    private static final int QUEUELIST = 9;
    private static final int QUEUESONG = 10;
    private static final int ADDLIBRARY = 11;
    private static final int REMOVELIBRARY = 12;
    private static final int CONNECT = 13;
    private static final int SHUFFLE = 14;

    private static final int ALLREADY_CONNECTED = 21421;

    private SpotifyAppRemote mSpotifyAppRemote;
    private String mClientID, mRedirectURI;
    private ConnectionParams mConnectionParams;
    private Track mTrack;

    private CallbackContext mCallbackContext;
    private int mQPosition = 0; // One is for iterating through songs that we want to queue
    private int currentPosInPlay = 0;
    private String mQFirstTrack = "";

    private PluginResult mResult;
    private boolean firstConnect = true;
    private PlayerState mPlayerState;

    private boolean movingForward = true;
    private boolean leftQList = false;
    private ArrayList<String> mQueueList = new ArrayList<>();
    private ArrayList<String> playedSongsStack;
    private String mTrackUri;
    private int mCurrentCase;

    @Override
    public boolean execute(String action, JSONArray args, CallbackContext callbackContext) throws JSONException {
        boolean returnThis;
        this.mCallbackContext = callbackContext;
        this.mResult = new PluginResult(PluginResult.Status.NO_RESULT);
        switch (action) {
            case "connectToRemote":
                if(this.mRedirectURI==null){
                    this.mRedirectURI = args.getString(0);
                    this.mClientID = args.getString(1);
                }
                this.connect();
                returnThis = true;
                break;
            case "builtInAuthConnect":
                this.mCurrentCase = CONNECT;
                if(this.mRedirectURI==null){
                    this.mRedirectURI = args.getString(0);
                    this.mClientID = args.getString(1);
                }
                Log.v("built in auth", "called");
                this.builtInAuth(mRedirectURI, mClientID , this.mCallbackContext);
                returnThis = true;
                break;
            case "disconnectRemote":
                this.disconnectRemote();
                returnThis = true;
                break;
            case "play":

                this.mTrackUri = args.getString(0);
                this.mCurrentCase = PLAY;
                this.issueToRemote();
                returnThis = true;
                break;

            case "pause":
                this.mCurrentCase = PAUSE;
                this.issueToRemote();
                returnThis = true;
                break;

            case "resume":
                this.mCurrentCase = RESUME;
                this.issueToRemote();
                returnThis = true;
                break;

            case "skipPrevious":

                this.mCurrentCase = SKIPBACK;
                this.issueToRemote();
                returnThis = true;
                break;

            case "skipNext":

                this.mCurrentCase = SKIPNEXT;
                this.issueToRemote();
                returnThis = true;
                break;

            case "queueSong":
                this.mTrackUri = args.getString(0);
                this.mCurrentCase = QUEUESONG;
                this.issueToRemote();
                returnThis = true;
                break;

            case "queueTracks":

                ArrayList tracksList = this.convertJSONArrayToList(args.getJSONArray(0));
                this.mCurrentCase = QUEUELIST;
                this.mQueueList = tracksList;
                this.leftQList = false;
                this.issueToRemote();
                returnThis = true;
                break;

            case "queueTracksNPlay":
                this.leftQList = false;
                ArrayList tracksListPlay = this.convertJSONArrayToList(args.getJSONArray(0));
                this.mCurrentCase = PLAYQUEUE;
                this.mQueueList = tracksListPlay;
                this.issueToRemote();
                returnThis = true;
                break;

            case "getCurrentTrack":
                this.returnCurrentTrack();
                returnThis = true;
                break;

            case "addToLibrary":

                this.mTrackUri = args.getString(0);
                this.mCurrentCase = ADDLIBRARY;
                this.issueToRemote();
                returnThis = true;
                break;

            case "removeFromLibrary":

                this.mTrackUri = args.getString(0);
                this.mCurrentCase = REMOVELIBRARY;
                this.issueToRemote();
                returnThis = true;
                break;

            case "openSearch":
                this.openSearch();
                returnThis = true;
                break;

            case "openUserPlaylist":
                this.openUserPlaylist();
                returnThis = true;
                break;
            case "openArtist":
                this.openArtist(args.getString(0));
                returnThis = true;
                break;
            case "openAlbum":
                this.openAlbum(args.getString(0));
                returnThis = true;
                break;
            case "openSong":
                Log.v("Args:: ", args.getString(0));
                this.openSong(args.getString(0));
                returnThis = true;
                break;

            case "shuffle":
                this.mCurrentCase = SHUFFLE;
                this.issueToRemote();
                returnThis = true;
                break;

            default:
                returnThis = false;
                break;

        }

        return returnThis;
    }

    @Override
    public void onActivityResult(int requestCode, int resultCode, Intent intent) {
        Log.v("^R^ best thing happened",  "");
        super.onActivityResult(requestCode, resultCode, intent);
        Log.v("^R^ Request Code: ", requestCode+"");
        // Check if result comes from the correct activity
        if (requestCode == 1330) {
            AuthorizationResponse response = AuthorizationClient.getResponse(resultCode, intent);
            Log.v("^R^ response code: ", response.getAccessToken());
            Log.v("^R^ expiration: ", response.getExpiresIn()+"");
            switch (response.getType()) {

                // Response was successful and contains auth token
                case TOKEN:
                    JSONObject r = new JSONObject();
                    try {
                        r.put("accessToken", response.getAccessToken());
                        r.put("expiresAt", response.getExpiresIn());
                        mResult = new PluginResult(PluginResult.Status.OK, r);
                        // extremely important, this has to be called each time the track changes,
                        // calling it once means call it allow only one more time.
                        this.mResult.setKeepCallback(true);
                        mCallbackContext.sendPluginResult(mResult);
                    } catch (JSONException e) {
                        e.printStackTrace();
                    }
                    // Handle successful response
                    this.connect();
                    break;

                // Auth flow returned an error
                case ERROR:
                    // Handle error response
                    break;

                // Most likely auth flow was cancelled
                default:
                    // Handle other cases
            }
        }
    }


    private void builtInAuth(String redirectURI, String clientID, CallbackContext callbackContext) {
        // this.connect();
        int REQUEST_CODE = 1330;


        try{
//            Log.v("^R^ parent: ", cordova.getContext().toString());
//            cordova.startActivityForResult();
            cordova.setActivityResultCallback (this);
            if(this.mSpotifyAppRemote == null){
                AuthorizationRequest.Builder builder =
                        new AuthorizationRequest.Builder(this.mClientID, AuthorizationResponse.Type.TOKEN, "https://io.ionic.latitunes");

                builder.setScopes(new String[]{"user-read-recently-played", "user-follow-read", "user-library-read", "user-library-modify"});
                Log.v("^R^ scopes added:", "");
                AuthorizationRequest request = builder.build();
                AuthorizationClient.openLoginActivity(cordova.getActivity(), REQUEST_CODE, request);
            }else if(!this.mSpotifyAppRemote.isConnected()){
                this.connect();
            }


        }catch(Error e){
            Log.d("^R^ Error,", e.getMessage());
        }

    }

    @Override
    public void onRestoreStateForActivityResult(Bundle state, CallbackContext callbackContext) {
        Log.v("^R^ RestoreState", "");
        super.onRestoreStateForActivityResult(state, callbackContext);
    }

    // Always call this method first to use the remote!!!!!!!!!!!!!!!!!!!!!
    // First step in using the remote, needs to get an instance of the spotify
    // remote and initialize the local member variable
    private void connect() {

        if (this.mSpotifyAppRemote != null) {
            if (this.mSpotifyAppRemote.isConnected() == false) {
                this.callConnectToSpotify();
            }else{
                this.returnCurrentTrack();
            }
        } else {
            this.callConnectToSpotify();
        }

    }
    public void callConnectToSpotify(){
        ConnectionParams connectionParams = new ConnectionParams.Builder(this.mClientID)
                .setRedirectUri(this.mRedirectURI)
                .showAuthView(true)
                .build();
        SpotifyAppRemote.connect(cordova.getActivity(), connectionParams,
                new Connector.ConnectionListener() {

                    public void onConnected(SpotifyAppRemote spotifyAppRemote) {
                        mSpotifyAppRemote = spotifyAppRemote;
                        Log.d("MainActivity", "Connected! Yay!");

                        firstConnect = false;
                        subscribeToPlayerEventChanges();
                        issueToRemote();
                    }

                    public void onFailure(Throwable throwable) {
                        Log.e("MyActivity", throwable.getMessage(), throwable);

                        // Something went wrong when attempting to connect! Handle errors here
                    }
                });
    }

    private void handleCommandType(int type) {
        switch (type) {
            case PLAY:
                this.play(this.mTrackUri, this.mCallbackContext);
                break;
            case PAUSE:
                this.pause();
                break;
            case RESUME:
                this.resume();
                break;
            case SKIPNEXT:
                this.skipNext();
                break;
            case SKIPBACK:
                this.skipPrevious();
                break;
            case PLAYQUEUE:
                this.queueTracksNPlay(this.mQueueList);
                break;
            case QUEUELIST:
                this.queueTracks(this.mQueueList);
                break;
            case QUEUESONG:
                this.queueSong(this.mTrackUri);
                break;
            case ADDLIBRARY:
                this.addToLibrary(this.mTrackUri);
                break;
            case REMOVELIBRARY:
                this.removeFromLibrary(this.mTrackUri);
                break;
            case CONNECT:
                this.returnCurrentTrack();
                break;
            case SHUFFLE:

            default:
                this.returnCurrentTrack();
                break;

        }
    }

    private void issueToRemote() {
        if (this.mSpotifyAppRemote != null) {
            if (this.mSpotifyAppRemote.isConnected() == false) {
                this.connect();
            } else {
                CordovaSpotifyRemote.this.handleCommandType(CordovaSpotifyRemote.this.mCurrentCase);
            }
        } else {
            this.connect();
        }

    }

    // Play the given Spotify uri with specific behaviour for that streamtype - such
    // as initiating playback of an URI on the audio stream for alarms.
    private void play(String spotifyID, CallbackContext callbackContext) {

        this.mSpotifyAppRemote.getPlayerApi().play(spotifyID);

    }

    private void subscribeToPlayerEventChanges() {
        mResult.setKeepCallback(true);
        mSpotifyAppRemote.getPlayerApi()
                .subscribeToPlayerState()
                .setEventCallback(playerState -> {
                    CordovaSpotifyRemote.this.mPlayerState = playerState;

                    final Track track = playerState.track;
                    // Sometimes the callback returns the same track. Atleast thats whats been
                    // logged
                    // Make sure this doesn't cause redundant or unessecary changes.
                    CordovaSpotifyRemote.this.mTrack = track;

                    JSONObject r = CordovaSpotifyRemote.this.getTrackJsonObj(track, this.mPlayerState.isPaused);
                    mResult = new PluginResult(PluginResult.Status.OK, r);
                    // extremely important, this has to be called each time the track changes,
                    // calling it once means call it allow only one more time.
                    this.mResult.setKeepCallback(true);
                    mCallbackContext.sendPluginResult(mResult);
                    if (track.uri != null) {
                        if (!this.mTrack.uri.equals(track.uri)) {
                            // If its something in the queue we have to treat things differently
                            if (this.checkInQueuedList(track.uri)) {
                                this.currentPosInPlay = this.mQueueList.indexOf(track.uri);
                            } else {
                                this.leftQList = true;
                                // Natural way.
                                if (this.movingForward)
                                    ++this.currentPosInPlay;
                                    // In the user previously had decided to skip to the previous song
                                    // We just need to decrement instead and set things back to the natural flow.
                                else {
                                    --this.currentPosInPlay;
                                    this.movingForward = true;
                                }
                            }

                        }

                    }

                });
    }

    private JSONObject getTrackJsonObj(Track track, boolean isPaused) {
        JSONObject r = new JSONObject();
        try {
            r.put("title", track.name);
        } catch (JSONException e) {
            e.printStackTrace();
        }
        try {
            r.put("artist", track.artist.name);
        } catch (JSONException e) {
            e.printStackTrace();
        }
        try {
            String imageURL = track.imageUri.raw.replaceAll("spotify:image:", "https://i.scdn.co/image/");
            r.put("imageUri", imageURL);
        } catch (JSONException e) {
            e.printStackTrace();
        }
        try {
            r.put("artistUri", track.artist.uri);
        } catch (JSONException e) {
        }
        try {
            r.put("albumUri", track.album.uri);
        } catch (JSONException e) {
        }
        try {
            r.put("trackUri", track.uri);
        } catch (JSONException e) {
        }
        try {
            r.put("album", track.album);
        } catch (JSONException e) {
        }
        try {
            r.put("isPaused", String.valueOf(isPaused));
        } catch (JSONException e) {
        }
        return r;
    }

    private void queueSong(String spotifyID) {
        this.mSpotifyAppRemote.getPlayerApi().queue(spotifyID);
    }

    private void queueTracksNPlay(ArrayList<String> spotifyIDs) {
        this.mQPosition = 0;
        this.mQFirstTrack = spotifyIDs.get(0);
        this.mSpotifyAppRemote.getPlayerApi().play(this.mQFirstTrack)
                .setResultCallback(playerState -> {
                    this.waitForPlayEventForQing();
                }).setErrorCallback(throwable -> {
                    Log.e("Error::", "while trying to queue");
                });
    }

    private void waitForPlayEventForQing() {
        this.mSpotifyAppRemote.getPlayerApi().subscribeToPlayerState().setEventCallback(playerState -> {

            if (playerState.track.uri.equals(this.mQFirstTrack)) {
                this.mTrack = playerState.track;
                // Now that the frist track has officially started playing we can now start
                // queueing.
                this.queueTracks(this.mQueueList);
            }

        }).setErrorCallback(throwable -> {
            Log.e("Error::", "while trying to queue");
        });
    }

    private void queueTracks(ArrayList<String> spotifyIDs) {
        this.mQueueList = spotifyIDs;
        // Now that the frist track has officially started playing we can now start
        // queueing.
        this.mSpotifyAppRemote.getPlayerApi()
                .getPlayerState()
                .setResultCallback(playerState -> {
                    this.mQPosition += 1;
                    // This prevents infinite recursion. What is happening is that we are always
                    // waiting
                    // To get a sign from spotify that it queued something successfully.
                    // This callback happens everytime an event happens inside spotify.
                    // What we are doing is telling it to keep queuing until all the songs are done
                    // But also to wait and ensure the last job was completed successfully before
                    // the next song is queued
                    if (this.mQueueList.size() > this.mQPosition) {
                        this.queueNCallback();
                    }
                }).setErrorCallback(throwable -> {
                    // In the case of an error, we are telling it to call a function that will
                    // restart this process.
                    if (this.mQueueList.size() > this.mQPosition) {
                        this.queueNCallback();
                    }
                });
    }

    private void queueNCallback() {
        this.mSpotifyAppRemote.getPlayerApi()
                .queue(this.mQueueList.get(this.mQPosition))
                .setResultCallback(playerState -> {
                    queueTracks(this.mQueueList);
                });

    }

    // Pause playback if it's currently playing Introduced in
    // com.spotify.protocol.client.RequiredFeatures#FEATURES_V1
    private void pause() {
        if (this.mPlayerState.isPaused) {
            this.mSpotifyAppRemote.getPlayerApi().resume();
        } else {
            this.mSpotifyAppRemote.getPlayerApi().pause();
        }

    }

    // Resume playback if it's currently paused Introduced in
    // com.spotify.protocol.client.RequiredFeatures#FEATURES_V1
    private void resume() {
        if (this.mPlayerState.isPaused) {
            this.mSpotifyAppRemote.getPlayerApi().resume();
        } else {
            this.mSpotifyAppRemote.getPlayerApi().pause();
        }
    }

    // Skip to the next track in the currently playing context if there is one.
    private void skipNext() {
        if (this.currentPosInPlay < this.mQueueList.size() && this.leftQList) {
            this.play(this.mQueueList.get(++this.currentPosInPlay), this.mCallbackContext);
        } else
            this.mSpotifyAppRemote.getPlayerApi().skipNext();
    }

    // Restart the current track or, if current track has played for less than 3
    // seconds, skip to the previous track in the currently playing context
    private void skipPrevious() {
        this.movingForward = false;
        if (this.currentPosInPlay < this.mQueueList.size() && this.currentPosInPlay > 0) {
            this.play(this.mQueueList.get(--this.currentPosInPlay), this.mCallbackContext);
        } else {
            this.mSpotifyAppRemote.getPlayerApi().skipPrevious();
        }
    }

    private void shuffle() {
        // this.mSpotifyAppRemote.shuffle();
    }

    private void disconnectRemote() {
        this.mResult.setKeepCallback(false);
        SpotifyAppRemote.disconnect(this.mSpotifyAppRemote);

    }

    private void addToLibrary(String id) {
        this.mSpotifyAppRemote.getUserApi().addToLibrary("spotify:track:" + id);
    }

    private void removeFromLibrary(String id) {
        this.mSpotifyAppRemote.getUserApi().removeFromLibrary("spotify:track:" + id);
    }

    private void openSearch() {
        Intent intent = new Intent(Intent.ACTION_VIEW);
        intent.setData(Uri.parse("spotify:search"));
        // Links us back to latitunes
        intent.putExtra(Intent.EXTRA_REFERRER,
                Uri.parse("android-app://" + cordova.getActivity().getPackageName()));
        cordova.getActivity().startActivity(intent);
    }

    private void openUserPlaylist() {
        Intent intent = new Intent(Intent.ACTION_VIEW);
        intent.setData(Uri.parse("spotify:playlists"));
        intent.putExtra(Intent.EXTRA_REFERRER,
                Uri.parse("android-app://" + cordova.getActivity().getPackageName()));
        cordova.getActivity().startActivity(intent);
    }

    private void openSong(String songUrl) {
        Intent intent = new Intent(Intent.ACTION_VIEW);
        intent.setData(Uri.parse(songUrl));
        // Links us back to latitunes
        intent.putExtra(Intent.EXTRA_REFERRER,
                Uri.parse("android-app://" + cordova.getActivity().getPackageName()));
        cordova.getActivity().startActivity(intent);
    }

    private void openArtist(String artistUrl) {
        Intent intent = new Intent(Intent.ACTION_VIEW);
        intent.setData(Uri.parse(this.mTrack.artist.uri));
        // Links us back to latitunes
        intent.putExtra(Intent.EXTRA_REFERRER,
                Uri.parse("android-app://" + cordova.getActivity().getPackageName()));
        cordova.getActivity().startActivity(intent);
    }

    private void openAlbum(String albumUrl) {
        Intent intent = new Intent(Intent.ACTION_VIEW);
        intent.setData(Uri.parse(this.mTrack.album.uri));
        // Links us back to latitunes
        intent.putExtra(Intent.EXTRA_REFERRER,
                Uri.parse("android-app://" + cordova.getActivity().getPackageName()));
        cordova.getActivity().startActivity(intent);
    }

    public ArrayList<String> convertJSONArrayToList(JSONArray jArray) {
        ArrayList<String> listdata = new ArrayList<String>();
        if (jArray != null) {
            for (int i = 0; i < jArray.length(); ++i) {
                try {
                    listdata.add(jArray.getString(i));
                } catch (JSONException e) {
                    e.printStackTrace();
                }
            }
        }
        return listdata;
    }

    public void returnCurrentTrack() {
        if (this.mTrack == null) {
            JSONObject r = this.getTrackJsonObj(this.mTrack, this.mPlayerState.isPaused);
            // SETING KEEPCALLBACK allows the javascript end to keep track of which track is
            // currently playing on Spotify
            mResult = new PluginResult(PluginResult.Status.OK, r);
            mCallbackContext.sendPluginResult(mResult);
        }
    }

    public boolean checkInQueuedList(String trackId) {
        if (this.mQueueList.contains(trackId))
            return true;
        else
            return false;
    }
}
